//*******************************************************
//
//	MP3 Player in SEGA Visual Memory
//
//	シリアルフラッシュROM関係
//
//	ATmega1284p
//		efuse:0xFD	BOD 2.7V
//		hfuse:0xD1	JTAG禁止, EEPROM消去保護
//		lfuse:0xE2	内蔵CR発振 8MHz
//
//	Compiler:	AVR-GCC		4.3.5
//				AVR-LIBC	1.6.8
//				binutils	2.20
//
//	2012.5.25	KAWAKAMI Yukio
//*******************************************************

#include	<avr/io.h>
#include	<avr/interrupt.h>
#include	<avr/pgmspace.h>
#include	<avr/eeprom.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>

#include	"main.h"
#include	"rom.h"
#include	"ff.h"
#include	"lcd.h"


#define		MCS_ON		PORTD &= ~(1<<MCS)
#define		MCS_OFF		PORTD |= (1<<MCS)
#define		SPI_WAIT	while(!(SPSR & (1<<SPIF)))

//*******************************************************
//	リソース番号からROMアドレスだけ取得
//*******************************************************
ulong resource_adrs(uchar n){
	ulong	adrs;
	ulong	size;

	if (resource_adrs_size(n, &adrs, &size))	return 0;

	return adrs;
}


//*******************************************************
//	リソース番号からROMアドレスとサイズ取得
//*******************************************************
char resource_adrs_size(uchar n, ulong *a, ulong *s){
	int		i;
	ulong	adrs;
	ulong	size;
	ulong	table;

	if (n == 0)	return 1;

	adrs = ADRS_RESOURCE + 256 * 4;
	table = ADRS_RESOURCE + 4;
	for (i = 1; i < n; i++){
		ROM_read(table, (uchar *)&size, 4);
		adrs += size;
		table += 4;
	}
	ROM_read(table, (uchar *)&size, 4);
	if ((size == 0)||(size == 0xFFFFFFFF))	return 1;

	*a = adrs;
	*s = size;
	return 0;
}


//*******************************************************
//	リソースファイル読み込み
//*******************************************************
void load_resource(void){
	FILINFO	info;
	WORD	fdate, ftime;
	UINT	n;
	FIL		file;
	ulong		rom;
	uchar		buff[256];

	// タイムスタンプチェック
	if (f_stat(RESOURCE_FNAME, &info) != FR_OK)	return;
	ROM_read(ADRS_RESOURCE, (uchar *)&ftime, 2);
	ROM_read(ADRS_RESOURCE + 2, (uchar *)&fdate, 2);
	if ((fdate == info.fdate)&&(ftime == info.ftime))	return;

	ROM_unprotect_all_sectors();	// プロテクト解除

	print_P(PSTR("\nResource Erase "));
	for (n = 0; n <= (info.fsize >> 16); n++){
		WDR;
		ROM_sector_erase(n + (ADRS_RESOURCE >> 16));
		putch('.');
	}
	print_P(PSTR("\nResource Load  "));
	if (f_open(&file, RESOURCE_FNAME, FA_READ) != FR_OK)	return;

	rom = ADRS_RESOURCE;
	if (f_read(&file, buff, 256, &n) != FR_OK)	return;
	// 最初の 4bytesはタイムスタンプを書く
	buff[0] = info.ftime;
	buff[1] = info.ftime >> 8;
	buff[2] = info.fdate;
	buff[3] = info.fdate >> 8;
	ROM_write_block(rom, buff, 256);
	rom += 256;

	while(f_read(&file, buff, 256, &n) == FR_OK){
		if (n == 0)	break;
		WDR;
		ROM_write_block(rom, buff, 256);
		rom += 256;
		if (!(rom & 65535)){
			putch('.');
		}
		memset(buff, 0xFF, 256);
	}

	print_P(PSTR("\nOK"));	// 終了
}


//*********************************************************
//	M25P20/AT26DF081A/SST25VF032B 書き込み終了待ち
//*********************************************************
void ROM_wait(void){

	// 書き込み終了待ち
	while(ROM_status() & 1){
		NOP;
	}
}


//*********************************************************
//	M25P20/AT26DF081A/SST25VF032B 読み込み
//*********************************************************
void ROM_read(ulong adrs, uchar *buff, int n){

	MCS_ON;
	SPI_out(0x03);
	SPI_out((uchar)(adrs >> 16));
	SPI_out((uchar)(adrs >> 8));
	SPI_out((uchar)adrs);

	SPDR = 0xFF;	// SPI転送開始
	while(n > 1){
		uchar	data;
		SPI_WAIT;
		data = SPDR;
		SPDR = 0xFF;
		*buff++ = data;
		--n;
	}
	SPI_WAIT;
	*buff = SPDR;
	MCS_OFF;
}


//*********************************************************
//	SST25VF032B 書き込み
//*********************************************************
void ROM_write(ulong adrs, uchar *buff, int n){

	// 念のため書き込み終了待ち
	ROM_wait();

	while(n > 0){

		MCS_ON;
		SPI_out(0x06);		// WREN
		MCS_OFF;

		MCS_ON;
		SPI_out(0x02);	// Byte-Program
		SPI_out((uchar)(adrs >> 16));
		SPI_out((uchar)(adrs >> 8));
		SPI_out((uchar)adrs);
		SPI_out(*buff++);
		MCS_OFF;
		adrs++;
		--n;

		ROM_wait();
	}
}


//*********************************************************
//	SST25VF032B ブロック書き込み
//*********************************************************
void ROM_write_block(ulong adrs, uchar *buff, int n){

	// 念のため書き込み終了待ち
	ROM_wait();

	MCS_ON;
	SPI_out(0x06);		// WREN
	MCS_OFF;

	MCS_ON;
	SPI_out(0xAD);	// AAI write
	SPI_out((uchar)(adrs >> 16));
	SPI_out((uchar)(adrs >> 8));
	SPI_out((uchar)adrs);
	SPI_out(*buff++);
	SPI_out(*buff++);
	MCS_OFF;

	n -= 2;

	while(n > 0){
		ROM_wait();

		MCS_ON;
		SPI_out(0xAD);	// AAI write
		SPI_out(*buff++);
		SPI_out(*buff++);
		MCS_OFF;
		n -= 2;
	}

	ROM_wait();

	MCS_ON;
	SPI_out(0x04);		// WRDI
	MCS_OFF;
}


//*********************************************************
//	M25P20/AT26DF081A/SST25VF032B Status Register 読み込み
//*********************************************************
uchar ROM_status(void){
	uchar	stat;

	MCS_ON;
	SPI_out(0x05);
	stat = SPI_in();
	MCS_OFF;

	return stat;
}


//*********************************************************
//	SST25VF032B セクタ消去
//	１セクタは 64KB
//*********************************************************
void ROM_sector_erase(uchar sector){

	ROM_wait();	// 消去待ち

	MCS_ON;
	SPI_out(0x06);	// 書き込み許可。１回書き込むと自動解除されるらしい
	MCS_OFF;

	MCS_ON;
	SPI_out(0xD8);
	SPI_out(sector);
	SPI_out(0);
	SPI_out(0);
	MCS_OFF;

	ROM_wait();	// 消去待ち
}


//*********************************************************
//	SST25VF032B プロテクト解除
//*********************************************************
void ROM_unprotect_all_sectors(void){

	MCS_ON;
	SPI_out(0x06);	// 書き込み許可。１回書き込むと自動解除されるらしい
	MCS_OFF;

	MCS_ON;
	SPI_out(0x01);	// WRSR
	SPI_out(0);
	MCS_OFF;
}


